/***************************************
	Auteur : Pierre Aubert
	Mail : aubertp7@gmail.com
	Licence : CeCILL-C
****************************************/

#include <map>
#include "H5Cpp.h"

#include "string_utils.h"
#include "convertToString.h"

#include "OptionParser.h"

#include "POffsetFinder.h"

#include "PDataCol.h"

using namespace std;

///Map of data column
typedef std::map<std::string, PDataCol> PMapDataCol;

///Create the OptionParser of this program
/**	@return OptionParser of this program
*/
OptionParser createOptionParser(){
	OptionParser parser(true, __PROGRAM_VERSION__);
	parser.setExampleLongOption("phoenix_binary_analyzer  --inputraw=rawMessage.h5 --inputfull=fullEvent.h5");
	parser.setExampleShortOption("phoenix_binary_analyzer  -i rawMessage.h5 -f fullEvent.h5");
	
	parser.addOption("inputraw", "i", OptionType::FILENAME, true, "input file of raw message file");
	parser.addOption("inputfull", "f", OptionType::FILENAME, true, "input file of fully decoded messages");
	return parser;
}

///Get the name of the main DataSet in the file
/**	@param file : HDF5 file to be used
 * 	@return name of the first DataSet
*/
std::string getDataSetName(const H5::H5File & file){
	hsize_t nbObj = file.getNumObjs();
	if(nbObj == 0lu){
		std::cerr << "getDataSetName : no object in file" << std::endl;
		return "";
	}
	if(nbObj != 1lu){
		std::cerr << "getDataSetName : expect only one object in file" << std::endl;
		return "";
	}
	H5G_obj_t objType = file.getObjTypeByIdx(0lu);
	if(objType != H5G_DATASET){
		std::cerr << "getDataSetName : expect only one DataSet in file" << std::endl;
		return "";
	}
	std::string nameDataSet(file.getObjnameByIdx(0lu));
	return nameDataSet;
}

///Load a PDataCol from a DataSet
/**	@param[out] dataCol : PDataCol to be loaded with one column of a DataSet
 * 	@param dataset : full DataSet
 * 	@param nbRowDataSet : number of rows of the DataSet
 * 	@param colName : name of the current column of the DataSet
 * 	@param colCompType : type of the column of the DataSet
 * 	@return true on success, false otherwise
*/
bool loadPDataColFromDataSetCol(PDataCol & dataCol, const H5::DataSet & dataset, size_t nbRowDataSet,
				const std::string & colName, const H5::CompType & colCompType)
{
	dataCol.setName(colName);
	size_t rowSize(colCompType.getSize());
	dataCol.setRowSize(rowSize);
	std::vector<char> & tabData = dataCol.getTabData();
	tabData.resize(rowSize*nbRowDataSet);
	//Create the column type we want to load
	H5::CompType loadColType(rowSize);
	loadColType.insertMember(colName, 0, colCompType);
	//Then load it
	dataset.read(tabData.data(), loadColType);
	return true;
}

///Load all the data columns of the given file
/**	@param[out] mapDataCol : PMapDataCol to be updated
 * 	@param file : file to be read
 * 	@return true on success, false otherwise
*/
bool loadMapDataCol(PMapDataCol & mapDataCol, const H5::H5File & file){
	//Find the single DataSet at the top of the file
	std::string nameDataSet(getDataSetName(file));
	if(nameDataSet == ""){return false;}
	//Open this DataSet
	H5::DataSet dataset = file.openDataSet(nameDataSet);
	//Get the number of rows of the DataSet
	H5::DataSpace dataSpace = dataset.getSpace();
	size_t nbRowDataSet(dataSpace.getSimpleExtentNpoints());
	//Find the different columns of this DataSet
	H5::CompType compType = dataset.getCompType();
	int nbMembers = compType.getNmembers();		//Number of columns in the DataSet
	if(nbMembers <= 0){
		std::cerr << "loadMapDataCol : the number of columns of the DataSet has to be >= 1" << std::endl;
		return false;
	}
	//Read each column and create the corresponding PDataCol
	for(int i(0); i < nbMembers; ++i){
		H5::CompType colCompType = compType.getMemberCompType(i);
		std::string colName(compType.getMemberName(i));
		PDataCol dataCol;
		if(loadPDataColFromDataSetCol(dataCol, dataset, nbRowDataSet, colName, colCompType)){
			mapDataCol[colName] = dataCol;
		}else{
			std::cerr << "loadMapDataCol : cannot load column '"<<colName<<"' of DataSet" << std::endl;
			return false;
		}
	}
	return true;
}

///Update the attributes to be found
/**	@param[out] finder : POffsetFinder to be updated with DataSet column
 * 	@param[out] mapDataCol : map of all the columns of the DataSet
 * 	@param fileInputFull : input file of fully decoded messages
 * 	@return true on success, false otherwise
*/
bool updateAttributeToBeFound(POffsetFinder & finder, PMapDataCol & mapDataCol, const std::string & fileInputFull){
	H5::H5File fileFull(fileInputFull, H5F_ACC_RDONLY);
	if(!loadMapDataCol(mapDataCol, fileFull)){
		std::cerr << "updateAttributeToBeFound : cannot load columns of file '"<<fileInputFull<<"'" << std::endl;
		return false;
	}
	for(PMapDataCol::iterator it(mapDataCol.begin()); it != mapDataCol.end(); ++it){
		finder.addAttribute(it->second.getName(), it->second.getTabData().data(), it->second.getRowSize());
	}
	return true;
}

///Create the source from configuration file
/**	@param fileInputRaw : input file of raw message file
 * 	@param fileInputFull : input file of fully decoded messages
 * 	@return true on success, false otherwise
*/
bool computeAttributeOffset(const std::string & fileInputRaw, const std::string & fileInputFull){
	if(fileInputRaw == "" || fileInputFull == ""){return false;}
	POffsetFinder finder;
	PMapDataCol mapDataCol;
	if(!updateAttributeToBeFound(finder, mapDataCol, fileInputFull)){
		std::cerr << "computeAttributeOffset : cannot update attribute to be found with file '"<<fileInputFull<<"'" << std::endl;
		return false;
	}
	PMapDataCol mapRawData;
	H5::H5File fileRaw(fileInputRaw, H5F_ACC_RDONLY);
	if(!loadMapDataCol(mapRawData, fileRaw)){
		std::cerr << "updateAttributeToBeFound : cannot load raw data of file '"<<fileInputRaw<<"'" << std::endl;
		return false;
	}
	PMapDataCol::iterator it(mapRawData.begin());
	size_t messageSize(it->second.getRowSize());
	size_t nbMessage(it->second.getTabData().size()/messageSize);
	finder.computeOffset(it->second.getTabData().data(), messageSize, nbMessage);
	finder.print();
	
	return true;
}

int main(int argc, char** argv){
	OptionParser parser = createOptionParser();
	parser.parseArgument(argc, argv);
	
	const OptionMode & defaultMode = parser.getDefaultMode();
	std::string fileInputRaw(""), fileInputFull("");
	defaultMode.getValue(fileInputRaw, "inputraw");
	defaultMode.getValue(fileInputFull, "inputfull");
	
	return computeAttributeOffset(fileInputRaw, fileInputFull);
}


